/*
 * fsrewrite.c
 * Copyright (C) 2005-2006 Robin KAY. All rights reserved.
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only
 * (the "License").  You may not use this file except in compliance
 * with the License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <limits.h>
#include <unistd.h>

#include "fsrewrite.h"

#define STRBUF_INCR 128
#define STRBUF_INIT {NULL, 0, 0}

typedef struct {
  char *str;
  int slen;
  int blen;
} strbuf_t;

static const char usage_string[] =
  "Usage: %s [-b base_path_of_rule [-t target_path ...] ...] command\n";

static const char enctbl[] = ENCODING_TABLE_INIT;

void strbuf_append(strbuf_t *buf, const char *str, int slen)
{
  if (slen < 0) {
    slen = strlen(str);
  }
  if (buf->slen + slen + 1 > buf->blen) {
    buf->blen = ((((buf->slen + slen) / STRBUF_INCR) + 1) * STRBUF_INCR);
    buf->str = realloc(buf->str, buf->blen);
  }
  memcpy(buf->str + buf->slen, str, slen);
  buf->slen += slen;
  buf->str[buf->slen] = '\0';
}

void encode(char *buf, unsigned int x)
{
  int n, i;
  
  x &= 0xffff;
  if (x <= 0xf) {
    n = 0;
  }
  else if (x <= 0xff) {
    n = 1;
  }
  else if (x <= 0xfff) {
    n = 2;
  }
  else {
    n = 3;
  }
  
  for (i=0; i<=n; i++) {
    int value;
    
    value = (x >> (DATA_BITS * (n-i))) & DATA_MASK;
    if (i == 0) {
      value |= START_BIT;
    }
    if (i == n) {
      value |= STOP_BIT;
    }
    buf[i] = enctbl[value];
  }
  
  buf[i] = '\0';
}

int main(int argc, char **argv)
{
  strbuf_t rules_buf = STRBUF_INIT;
  strbuf_t pwd_buf = STRBUF_INIT;
  int seen_base_path, extra_path_flags, optarg_len;
  char *rules_env, *pwd_env;
  char chr, encbuf[5];
  
  strbuf_append(&rules_buf, RULES_ENV_NAME "=", -1);
  seen_base_path = 0;
  extra_path_flags = 0;
  
  while ((chr = getopt(argc, argv, "sSb:t:")) != -1) {
    switch (chr) {
      case 's': {
        extra_path_flags &= ~PATH_FLAGS_IGNORE_SYMS;
        break;
      }
      case 'S': {
        extra_path_flags |= PATH_FLAGS_IGNORE_SYMS;
        break;
      }
      case 'b': {
        seen_base_path = 1;
        if (optarg[0] != '/') {
          fprintf(stderr, "%s: Error: Paths must be absolute.\n", argv[0]);
          return EXIT_FAILURE;
        }
        optarg_len = strlen(optarg);
        encode(encbuf, PATH_FLAGS_BASE | extra_path_flags);
        strbuf_append(&rules_buf, encbuf, -1);
        encode(encbuf, optarg_len);
        strbuf_append(&rules_buf, encbuf, -1);
        strbuf_append(&rules_buf, optarg, optarg_len);
        strbuf_append(&rules_buf, encbuf, -1);
        break;
      }
      case 't': {
        if (!seen_base_path) {
          fprintf(stderr, usage_string, argv[0]);
          return EXIT_FAILURE;
        }         
        if (optarg[0] != '/') {
          fprintf(stderr, "%s: Error: Paths must be absolute.\n", argv[0]);
          return EXIT_FAILURE;
        }
        optarg_len = strlen(optarg);
        encode(encbuf, PATH_FLAGS_TARGET | extra_path_flags);
        strbuf_append(&rules_buf, encbuf, -1);
        encode(encbuf, optarg_len);
        strbuf_append(&rules_buf, encbuf, -1);
        strbuf_append(&rules_buf, optarg, optarg_len);
        strbuf_append(&rules_buf, encbuf, -1);
        break;
      }
      case '?': {
        fprintf(stderr, usage_string, argv[0]);
        return EXIT_FAILURE;
      }
    }
  }
  
  if (optind >= argc) {
    fprintf(stderr, usage_string, argv[0]);
    return EXIT_FAILURE;
  }
  
  if (!seen_base_path) {
    fprintf(stderr, "%s: Warning: No rules have been specified.\n", argv[0]);
  }
  
  // Set rules environment variables
  rules_env = getenv(RULES_ENV_NAME);
  if (rules_env) {
    strbuf_append(&rules_buf, rules_env, -1);
  }
  putenv(rules_buf.str);
  
  // Set pwd environment variables
  strbuf_append(&pwd_buf, PWD_ENV_NAME "=", -1);
  pwd_env = getenv(PWD_ENV_NAME);
  if (pwd_env) {
    strbuf_append(&pwd_buf, pwd_env, -1);
  }
  else {
    char buf[PATH_MAX];
    
    if (!getcwd(buf, PATH_MAX)) {
      return EXIT_FAILURE;
    }
    strbuf_append(&pwd_buf, buf, -1);
  }
  putenv(pwd_buf.str);
  
  putenv("LD_AUDIT_32=" AUDITOR_INST_PATH_32);
  putenv("LD_AUDIT_64=" AUDITOR_INST_PATH_64);
  putenv("LD_PRELOAD_32=" MALLORY_INST_PATH_32);
  putenv("LD_PRELOAD_64=" MALLORY_INST_PATH_64);
  
  execvp(argv[optind], &(argv[optind]));
  perror(argv[0]);
  return EXIT_FAILURE;
}
