/*
 * tracker.c
 * Copyright (C) 2005-2006 Robin KAY. All rights reserved.
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only
 * (the "License").  You may not use this file except in compliance
 * with the License.
 */

#include "compat.h"

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <pthread.h>

#include "fsrewrite.h"
#include "utils.h"
#include "mallory.h"
#include "tracker.h"

#ifdef _ILP32
#define LLSEEK lseek64
#define LGETDENTS libc_getdents64
#define LDIRENT dirent64_t
#define LOFFSET off64_t
#else
#define LLSEEK lseek
#define LGETDENTS libc_getdents
#define LDIRENT dirent_t
#define LOFFSET off_t
#endif

#define GETDENTS_BUFSIZE 4096

#define LDIRENT_EXTRA (sizeof(LDIRENT)-sizeof(dirent_t))

typedef struct {
  int ref_count;
  char vpath_buf[PATH_MAX];
  int vpath_len;
  void *dirents_stack;
  tree_node_t *root_dirent;
  tree_node_t *first_dirent;
  tree_node_t *cached_dirent;
  LOFFSET cached_offset, end_offset;  
} dir_rec_t;

typedef struct {
  int fd_hash;
  dir_rec_t *dir;
} dirfd_rec_t;

static mutex_t big_giant_lock;
static path_buf_t cwd;
static tree_node_t *dirfd_root = NULL;

static int hash_int(int key)
{
  // Robert Jenkins' 32-bit mix function
  key += (key << 12);
  key ^= (key >> 22);
  key += (key << 4);
  key ^= (key >> 9);
  key += (key << 10);
  key ^= (key >> 2);
  key += (key << 7);
  key ^= (key >> 12);
  
  return key;
}

static int dirfd_cmp_cb(void *l, void *r)
{
  dirfd_rec_t *left = (dirfd_rec_t *)l;
  dirfd_rec_t *right = (dirfd_rec_t *)r;
  
  return left->fd_hash - right->fd_hash;
}

static dirfd_rec_t *get_dirfd(int fd) {
  dirfd_rec_t rec;
  
  rec.fd_hash = hash_int(fd);
  return tree_op(&dirfd_root, &dirfd_cmp_cb, &rec, LOOKUP);
}

static void add_dirfd(int fd, dir_rec_t *dir)
{
  tree_node_t *node;
  dirfd_rec_t *dirfd;

  node = block_alloc();
  dirfd = HDR_TO(tree_node_t, node);
  dirfd->fd_hash = hash_int(fd);
  dirfd->dir = dir;
  dirfd = tree_op(&dirfd_root, &dirfd_cmp_cb, dirfd, INSERT);
  dir->ref_count++;
}

static int get_path(path_buf_t *wpath, int fd)
{
  dirfd_rec_t *dirfd;
  
  // Return the current working directory if AT_FDCWD is passed
  if (fd == AT_FDCWD) {
    MUTEX_LOCK(&big_giant_lock);
    path_copy(wpath, &cwd);
    MUTEX_UNLOCK(&big_giant_lock);
    return 0;
  }
  
  // Get the path from the record
  MUTEX_LOCK(&big_giant_lock);
  dirfd = get_dirfd(fd);
  if (dirfd) {
    dir_rec_t *dir = dirfd->dir;
    
    memcpy(wpath->buf, dir->vpath_buf, dir->vpath_len);
    wpath->len = dir->vpath_len;
    MUTEX_UNLOCK(&big_giant_lock);
    return 0;
  }
  else {
    MUTEX_UNLOCK(&big_giant_lock);
    errno = ENOTDIR;
    return -1;
  }
}

void tracker_init()
{
  const char *pwd_env;
  
  // Initialize path rewriting engine
  path_rewrite_init(libc_readlink);
  
  // Copy pwd environment variable into cwd buffer
  pwd_env = getenv(PWD_ENV_NAME);
  if (!pwd_env) {
    fprintf(stderr, "fsrewrite: " PWD_ENV_NAME " is not set.\n");
    abort();
  }
  if (!path_change(path_set(&cwd, "/"), pwd_env)) {
    fprintf(stderr, "fsrewrite: " PWD_ENV_NAME " is too long.\n");
    abort();
  }
  
  // Initialize block allocator
  alloc_init(
    MAX(RECLEN(tree_node_t) + sizeof(dirfd_rec_t), sizeof(dir_rec_t)),
    RECLEN(tree_node_t) + sizeof(dirent_t) + MAXNAMLEN);
}

int tracker_open_fd(int fd, const char *vpath)
{
  struct stat statbuf;
  int ret;
  dir_rec_t *dir;
  
  // Check that the file descriptor refers to a directory
  while (((ret = fstat(fd, &statbuf)) < 0) && (errno == EINTR)) {}
  if (ret < 0) {
    return -1;
  }
  if ((statbuf.st_mode & S_IFMT) != S_IFDIR) {
    return 0;
  }
  
  // Initialize new directory record
  MUTEX_LOCK(&big_giant_lock);
  dir = block_alloc();
  dir->ref_count = 0;
  dir->vpath_len = strlen(vpath);
  memcpy(dir->vpath_buf, vpath, dir->vpath_len);
  dir->dirents_stack = NULL;
  dir->root_dirent = NULL;
  dir->first_dirent = NULL;
  dir->cached_dirent = NULL;
  dir->cached_offset = 0;
  dir->end_offset = 0;
  
  // Initialize new descriptor record
  add_dirfd(fd, dir);
  MUTEX_UNLOCK(&big_giant_lock);
  
  return 0;
}

void tracker_dup_fd(int fd, int fd2)
{
  dirfd_rec_t *dirfd;
  
  // Get directory record
  MUTEX_LOCK(&big_giant_lock);
  dirfd = get_dirfd(fd);
  if (!dirfd) {
    MUTEX_UNLOCK(&big_giant_lock);
    return;
  }
  
  // Initialize new descriptor record
  add_dirfd(fd2, dirfd->dir);
  MUTEX_UNLOCK(&big_giant_lock);
  
  return;
}

void tracker_close_fd(int fd)
{
  dirfd_rec_t *dirfd;
  dir_rec_t *dir;
  
  // Get descriptor record
  MUTEX_LOCK(&big_giant_lock);
  dirfd = get_dirfd(fd);
  if (!dirfd) {
    MUTEX_UNLOCK(&big_giant_lock);
    return;
  }
  
  // Free directory free if last descriptor
  dir = dirfd->dir;
  if (!--dir->ref_count) {
    stack_free(dir->dirents_stack);
    block_free(dir);
  }
  
  // Free descriptor record
  tree_op(&dirfd_root, &dirfd_cmp_cb, dirfd, REMOVE);
  block_free(TO_HDR(tree_node_t, dirfd));
  MUTEX_UNLOCK(&big_giant_lock);
  
  return;
}

int tracker_rewrite(
  int fd,
  const char *path,
  int rflags,
  path_rewrite_cb_t rewrite_cb,
  tracker_failed_cb_t failed_cb,
  void *cb_data)
{
  path_buf_t wpath;
  
  // Get path to file descriptor
  if (get_path(&wpath, fd) < 0) {
    if (failed_cb) {
      return failed_cb(fd, path, cb_data);
    }
    else {
      errno = ENOTDIR;
      return -1;
    }
  }
  
  // Convert relative paths to absolute form
  if (!path_change(&wpath, path)) {
    return -1;
  }
  
  return path_rewrite(&wpath, rflags, rewrite_cb, cb_data);
}

/* -----
 * chdir
 * -----
 */

static int chdir_cb(
  const char *path, const char *vpath, int pflags, void *cb_data)
{
  int retval;
  
  retval = libc_chdir(path);
  if (retval >= 0) {
    path_set(&cwd, vpath);
  }
  return retval;
}

int tracker_chdir(const char *path)
{
  path_buf_t wpath;
  int retval;
  
  MUTEX_LOCK(&big_giant_lock);
  if (!path_change(path_copy(&wpath, &cwd), path)) {
    MUTEX_UNLOCK(&big_giant_lock);
    return -1;
  }
  retval = path_rewrite(&wpath, REWRITE_FLAGS_NONE, &chdir_cb, NULL);
  MUTEX_UNLOCK(&big_giant_lock);
  
  return retval;
}

/* ------
 * fchdir
 * ------
 */

int tracker_fchdir(int fd)
{
  path_buf_t wpath;
  int retval;
  
  if (get_path(&wpath, fd) < 0) {
    return -1;
  }
  retval = path_rewrite(&wpath, REWRITE_FLAGS_NONE, &chdir_cb, NULL);
  
  return retval;
}

/* ------
 * getcwd
 * ------
 */

char *tracker_getcwd(char *buf, size_t size)
{
  MUTEX_LOCK(&big_giant_lock);
  if (size == 0) {
    errno = EINVAL;
    return NULL;
  }
  
  if (size < cwd.len + 1) {
    errno = ERANGE;
    return NULL;
  }
  if (!buf) {
    buf = malloc(size);
    if (!buf) {
      errno = ENOMEM;
      return NULL;
    }
  }
  strcpy(buf, path_get_cstr(&cwd));
  MUTEX_UNLOCK(&big_giant_lock);
  return buf; 
}

/* --------
 * getdents
 * --------
 */

static int dirent_cmp_cb(void *l, void *r)
{
  LDIRENT *left = (LDIRENT *)l;
  LDIRENT *right = (LDIRENT *)r;
  
  return strcmp(left->d_name, right->d_name);
}

static int build_dirents_cb(
  const char *path, const char *vpath, int pflags, void *cb_data)
{
  dir_rec_t *dir = (dir_rec_t *)cb_data;
  char buf[GETDENTS_BUFSIZE];
  int fd, bufsize;
  
  while (((fd = libc_open(path, O_RDONLY, 0)) < 0) && (errno == EINTR)) {}
  if (fd < 0) {
    return -1;
  }
  do {
    while ((bufsize = LGETDENTS(fd, (LDIRENT *)buf, GETDENTS_BUFSIZE)) > 0) {
      char *p = buf;
      while (p < (buf + bufsize)) {
        LDIRENT *dirent;
        tree_node_t **nodep;
        
        dirent = (LDIRENT *)p;
        nodep = tree_op(&dir->root_dirent, &dirent_cmp_cb, dirent, TEST_INSERT);
        if (nodep) {
          tree_node_t *node;
          
          node = stack_alloc(
            dir->dirents_stack, RECLEN(tree_node_t) + dirent->d_reclen);
          if (!dir->dirents_stack) {
            dir->dirents_stack = node;
          }
          memcpy(HDR_TO(tree_node_t, node), dirent, dirent->d_reclen);
          tree_op(nodep, &dirent_cmp_cb, HDR_TO(tree_node_t, node), INSERT);
        }
        p += dirent->d_reclen;
      }
    }
  } while ((bufsize < 0) && (errno == EINTR));
  libc_close(fd);
  if (bufsize < 0) {
    return -1;
  }
  
  errno = ENOENT;
  return -1;
}

static int build_dirents(dir_rec_t *dir)
{
  tree_node_t *first, *node;
  path_buf_t wpath;
  LOFFSET off;
  
  // Free any previous tree
  stack_free(dir->dirents_stack);
  dir->dirents_stack = NULL;
  
  // Populate the tree with unique names
  dir->vpath_buf[dir->vpath_len] = '\0';
  path_set(&wpath, dir->vpath_buf);
  if ((path_rewrite(
        &wpath, REWRITE_FLAGS_NONE, &build_dirents_cb, dir) < 0) &&
      (errno != ENOENT)) {
    stack_free(dir->dirents_stack);
    dir->dirents_stack = NULL;
    return -1;
  }
  first = tree_linearize(dir->root_dirent);
  
  // Recalculate offsets
  node = first;
  off = 0;
  while (node) {
    LDIRENT *dirent = (LDIRENT *)HDR_TO(tree_node_t, node);
    dirent->d_off = ++off;
    node = node->next;
  }
  
  dir->first_dirent = first;
  dir->cached_dirent = first;
  dir->cached_offset = 0;
  dir->end_offset = off;
  return 0;
}

int tracker_getdents(int fd, void *buf, size_t bufsize
#ifdef _ILP32
  , int lf64
#endif
  )
{
  dirfd_rec_t *dirfd;
  dir_rec_t *dir;
  tree_node_t *node;
  LDIRENT *dirent;
  LOFFSET off;
  int nbytes;
  
  // Get directory record
  MUTEX_LOCK(&big_giant_lock);
  dirfd = get_dirfd(fd);
  if (!dirfd) {
    MUTEX_UNLOCK(&big_giant_lock);
    errno = ENOTDIR;
    return -1;
  }
  dir = dirfd->dir;
  
  // Build or rebuild the list of dirents
  off = LLSEEK(fd, 0, SEEK_CUR);
  if (!dir->dirents_stack || ((off == 0) && (dir->cached_offset != 0))) {
    if (build_dirents(dir) < 0) {
      MUTEX_UNLOCK(&big_giant_lock);
      return -1;
    }
  }
  
  // Return if past end of directory
  if (off > dir->end_offset) {
    MUTEX_UNLOCK(&big_giant_lock);
    errno = ENOENT;
    return -1;
  }
  
  // Selected starting node
  if (off < dir->cached_offset) {
    node = dir->first_dirent;
  }
  else {
    node = dir->cached_dirent;
  }
  
  // Write dirents into application buffer
  nbytes = 0;
#ifdef _ILP32
  if (!lf64) {
    while (node &&
           (dirent = (LDIRENT *)HDR_TO(tree_node_t, node)) &&
           (nbytes + dirent->d_reclen - LDIRENT_EXTRA <= bufsize)) {
      dirent_t *dst_dirent = (dirent_t *)((char *)buf + nbytes);
      unsigned short reclen = dirent->d_reclen - LDIRENT_EXTRA;
      
      dst_dirent->d_ino = (ino_t)dirent->d_ino;
      dst_dirent->d_off = (off_t)dirent->d_off;
      
      // Check if values overflow a 32-bit dirent structure
      if ((dst_dirent->d_ino != dirent->d_ino) ||
          (dst_dirent->d_off != dirent->d_off)) {
        MUTEX_UNLOCK(&big_giant_lock);
        errno = EOVERFLOW;
        return -1;
      }
      
      // Copy entry name
      dst_dirent->d_reclen = reclen;
      memcpy(
        (char *)dst_dirent + offsetof(dirent_t, d_name),
        (char *)dirent + offsetof(LDIRENT, d_name),
        dirent->d_reclen - offsetof(dirent_t, d_name));
      
      nbytes += reclen;
      off++;
      node = node->next;
    }
  }
  else
#endif
  while (node &&
         (dirent = (LDIRENT *)HDR_TO(tree_node_t, node)) &&
         (nbytes + dirent->d_reclen <= bufsize)) {
    memcpy((char *)buf + nbytes, dirent, dirent->d_reclen);
    nbytes += dirent->d_reclen;
    off++;
    node = node->next;
  }
  
  // Set direcotry offset and update cache
  LLSEEK(fd, off, SEEK_SET);
  dir->cached_dirent = node;
  dir->cached_offset = off;
  
  if (nbytes || !node) {
    MUTEX_UNLOCK(&big_giant_lock);
    return nbytes;
  }
  else {
    MUTEX_UNLOCK(&big_giant_lock);
    errno = EINVAL;
    return -1;
  }
}
