/*
 * utils.c
 * Copyright (C) 2005-2007 Robin KAY. All rights reserved.
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only
 * (the "License").  You may not use this file except in compliance
 * with the License.
 */

#include "compat.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/mman.h>
#include <sys/param.h>
#include <errno.h>
#include <limits.h>
#include <unistd.h>

#include "fsrewrite.h"
#include "utils.h"

#define BLOCKS_PER_MAPPING LONG_BIT
#define ALL_BLOCK_BITS ~(blocks_bits_t)0

typedef unsigned long blocks_bits_t;

struct mapping_s {
  struct mapping_s *next_mapping;
  struct mapping_s **ptr_to_this;
  blocks_bits_t in_use_bits;
};
typedef struct mapping_s mapping_t;

struct block_s {
  int block_idx;
};
typedef struct block_s block_t;

struct bstack_s {
  struct bstack_s *head;
  struct bstack_s *next;
  int off;
};
typedef struct bstack_s bstack_t;

static readlink_fptr_t readlink_fptr;
static const char *rules_start, *rules_end;
static int dectbl[UCHAR_MAX+1];
static const char enctbl[] = ENCODING_TABLE_INIT;
static size_t mapping_size, block_size;
static mapping_t *first_mapping = NULL;

path_buf_t *path_copy(path_buf_t *dpath, path_buf_t *spath)
{
  if ((!dpath) || (!spath)) {
    errno = EFAULT;
    return NULL;
  }
  
  dpath->len = spath->len;
  memcpy(dpath->buf, spath->buf, spath->len);
  return dpath;
}

static path_buf_t *append(path_buf_t *wpath, const char *path, int path_len)
{
  if (wpath->len + path_len > PATH_MAX - 1) {
    errno = ENAMETOOLONG;
    return NULL;
  }
  memcpy(wpath->buf + wpath->len, path, path_len);
  wpath->len += path_len;
  return wpath;
}

path_buf_t *path_set(path_buf_t *wpath, const char *path)
{
  if (!wpath || !path) {
    errno = EFAULT;
    return NULL;
  }
  wpath->len = 0;
  return append(wpath, path, strlen(path));
}

path_buf_t *path_change(path_buf_t *wpath, const char *path)
{
  const char *start_p, *cur_p;
  
  if (!wpath || !path) {
    errno = EFAULT;
    return NULL;
  }
  if (path[0] == '/') {
    if (!path_set(wpath, "/")) {
      return NULL;
    }
  }
  
  start_p = cur_p = path;
  for (;;) {
    if (*cur_p == '/' || *cur_p == '\0') {
      int len = cur_p - start_p;
      
      if (strncmp(start_p, ".", len) != 0) {
        if (wpath->len > 1) {
          if (!append(wpath, "/", 1)) {
            return NULL;
          }
        }
        if (!append(wpath, start_p, len)) {
          return NULL;
        }
      }
      if (*cur_p == '\0') {
        break;
      }
      while (*cur_p == '/') {
        cur_p++;
      }
      start_p = cur_p;
    }
    else {
      cur_p++;
    }
  }
  
  return wpath;
}

char *path_get_cstr(path_buf_t *wpath)
{
  if (!wpath) {
    errno = EFAULT;
    return NULL;
  }
  
  wpath->buf[wpath->len] = '\0';
  return wpath->buf;
}

void path_rewrite_init(readlink_fptr_t fptr)
{
  int i;

  // Save readlink callback for later use
  readlink_fptr = fptr;

  // Read the rules from an environment variable
  rules_start = getenv(RULES_ENV_NAME);
  if (!rules_start) {
    fprintf(stderr, "fsrewrite: " RULES_ENV_NAME " is not set.\n");
    abort();
  }
  rules_end = rules_start + strlen(rules_start);
  
  // Setup decoding table
  for (i=0; i<=UCHAR_MAX; i++) {
    dectbl[i] = -1;
  }
  for (i=0; i<sizeof(enctbl); i++) {
    dectbl[(int)enctbl[i]] = i;
  }
}

static int decode(int reverse, const char **pp)
{
  int code, delta, value, count;
  const char *p;
  
  p = *pp;
  delta = 0;
  value = 0;
  count = 0;
  do {
    // Decode character
    if (reverse) {
      if (p < rules_start) {
        return -1;
      }
      code = dectbl[(int)*(p--)];
    }
    else {
      if (p > rules_end) {
        return -1;
      }
      code = dectbl[(int)*(p++)];
    }
    if (code < 0) {
      return -1;
    }
    
    // Process control bits
    if (code & START_BIT) {
      delta++;
    }
    if (code & STOP_BIT) {
      delta--;
    }
    
    // Append data bits to value
    if (delta < 0) {
      value = value | ((code & DATA_MASK) << (DATA_BITS * count));
      count++;
    }
    else {
      value = (value << DATA_BITS) | (code & DATA_MASK);
    }
  } while (delta);
  *pp = p;
  
  return value;
}

static const char *next_path(
  path_buf_t *wpath, int reverse, const char *p, int *flagsp)
{
  int flags, slen;
  const char *p2;
  
  // Read flags if starting at the beginning 
  flags = 0;
  if (reverse) {
    p--;
  }
  else {
    flags = decode(reverse, &p);
    if (flags < 0) {
      return NULL;
    }
  }
  
  // Read the first length specifier encountered
  slen = decode(reverse, &p);
  if (slen < 0) {
    return NULL;
  }
  
  // Check string is within valid range
  if (reverse) {
    p2 = p - slen;
    if (p2 < rules_start) {
      return NULL;
    }
  }
  else {
    p2 = p + slen;
    if (p2 > rules_end) {
      return NULL;
    }
  }
  
  // Copy path string into path buffer
  if (wpath) {
    wpath->len = 0;
    if (!append(wpath, p, slen)) {
      return NULL;
    }
  }
  p = p2;
  
  // Check the second length specifier matches the first
  if (decode(reverse, &p) != slen) {
    return NULL;
  }
  
  // Read flags if reversing towards the beginning
  if (reverse) {
    flags = decode(reverse, &p);
    if (flags < 0) {
      return NULL;
    }
  }
  
  *flagsp = flags;
  return p;
}

static int path_rewrite_1(
  const char *path, path_rewrite_cb_t cb, void *cb_data)
{
  const char *curr_rule, *next_rule, *best_rule, *best_path_suffix;
  int pflags, target_pflags, best_path_suffix_len, best_depth, best_pflags;
  path_buf_t wpath;
  
  // Iterate through the rule set to find the closest matching path
  curr_rule = rules_start;
  best_rule = NULL;
  best_path_suffix = NULL;
  best_path_suffix_len = 0;
  best_depth = 0;
  best_pflags = 0;
  
  while ((next_rule = next_path(&wpath, 0, curr_rule, &pflags)) &&
         (pflags & PATH_FLAGS_BASE)) {
    const char *basep = path_get_cstr(&wpath);
    const char *pathp = path;
    int match, depth;
    
    // Attempt to match path against candidate base path in the rule set
    match = 1;
    depth = 0;
    for (;;) {
      if (*basep == '/' || *basep == '\0') {
        depth++;
        while (*basep == '/') {
          basep++;
        }
        while (*pathp == '/') {
          pathp++;
        }
        if (*basep == '\0') {
          break;
        }
      }
      else if (*basep == *pathp) {
        match = 1;
        basep++;
        pathp++;
      }
      else {
        match = 0;
        break;
      }
    }
    
    // Test if match is better than the previous best
    if (match) {
      if (depth > best_depth) {
        best_rule = next_rule;
        best_path_suffix = pathp - (*pathp ? 1 : 0);
        best_path_suffix_len = strlen(best_path_suffix);
        best_depth = depth;
        best_pflags = pflags;
      }
    }
    
    // Skip all the rule's target paths
    curr_rule = next_rule;
    while ((next_rule = next_path(&wpath, 0, curr_rule, &pflags)) &&
           (pflags & PATH_FLAGS_TARGET)) {
      curr_rule = next_rule;
    }
  }
  
  // If there's no matching rule then use the original path
  if (!best_rule) {
    return cb(path, path, 0, cb_data);
  }
  
  // Iterate through the rule's target paths until the callback succeeds.
  curr_rule = best_rule;
  while ((next_rule = next_path(&wpath, 0, curr_rule, &target_pflags)) &&
         (target_pflags & PATH_FLAGS_TARGET)) {
    int retval;
    
    // Attach the suffix of the matched base path to a candidate target path
    if (!append(&wpath, best_path_suffix, best_path_suffix_len)) {
      return -1;
    }
    
    // Make callback
    retval = cb(path_get_cstr(&wpath), path,
      PFLAGS_COMBINE(best_pflags, pflags), cb_data);
    if ((retval >= 0) || (errno != ENOENT)) {
      return retval;
    }
    
    curr_rule = next_rule;
  }
  
  errno = ENOENT;
  return -1;
}

static int follow_symlink_cb(
  const char *path, const char *vpath, int pflags, void *cb_data)
{
  path_buf_t *wpath = (path_buf_t *)cb_data;
  RDLNK_RETTYPE len;
  
  if (PFLAGS_OVERLAY(pflags) & PATH_FLAGS_IGNORE_SYMS) {
    errno = EINVAL;
    return -1;
  }
  else {
    len = readlink_fptr(path, wpath->buf, PATH_MAX);
    if (len >= 0) {
      wpath->len = len;
    }
    return (len < 0) ? -1 : 0;
  }
}

int path_rewrite(
  path_buf_t *wpath, int rflags, path_rewrite_cb_t cb, void *cb_data)
{
  char *start_p, *cur_p;
  int old_errno, symloop, noent_ok;
  
  start_p = cur_p = path_get_cstr(wpath);
  old_errno = errno;
  symloop = noent_ok = 0;
  do {
    char *last_p;
    
    // Advance to next path component seperator
    last_p = cur_p;
    while (*cur_p == '/') {cur_p++;}
    while ((*cur_p != '/') && (*cur_p != '\0')) {cur_p++;}
    
    // Handle special cases for the last path component
    if (*cur_p == '\0') {
      if (rflags & REWRITE_FLAGS_IGN_LAST_SYM) {
        break;
      }
      else if (rflags & REWRITE_FLAGS_IGN_LAST_NOENT) {
        noent_ok = 1;
      }
    }
    
    if (strncmp(last_p, "/..", 3) == 0) {
      char *src_p, *dst_p;
      
      // Remove link to parent from path
      src_p = dst_p = last_p;
      src_p += 3;
      if (dst_p - start_p > 0) {
        while (*dst_p == '/') {dst_p--;}
        while (*dst_p != '/') {dst_p--;}
      }
      memmove(dst_p, src_p, wpath->len - (src_p - start_p));
      wpath->len -= src_p - dst_p;
      if (wpath->len < 1) {
        wpath->buf[0] = '/';
        wpath->len = 1;
      }
      start_p = path_get_cstr(wpath);
      cur_p = dst_p;
    }
    else {
      path_buf_t lpath;
      char c;
      int retval;
      
      // Test if subset of path is a symlink
      c = *cur_p;
      *cur_p = '\0';
      retval = path_rewrite_1(start_p, &follow_symlink_cb, &lpath);
      *cur_p = c;
      
      // Process symlink
      if (!retval) {
        int suffix_len;
        
        if (++symloop > MAXSYMLINKS) {
          errno = ELOOP;
          return -1;
        }
        
        // Append suffix of path to symlink to symlink contents
        suffix_len = wpath->len - (cur_p - start_p);
        memcpy(lpath.buf + lpath.len, cur_p, suffix_len);
        lpath.len += suffix_len;
        
        // Remove last component from path to symlink
        wpath->len = last_p - start_p;
        if (wpath->len < 1) {
          wpath->len = 1;
        }
        
        // Change path
        path_change(wpath, path_get_cstr(&lpath));
        start_p = path_get_cstr(wpath);
        if (lpath.buf[0] == '/') {
          cur_p = start_p;
        }
        else {
          cur_p = last_p;
        }
      }
      else if ((errno != EINVAL) && (errno != ENOSYS) &&
               ((errno != ENOENT) || !noent_ok)) {
        return -1;
      }
    }
  } while (*cur_p != '\0');
  errno = old_errno;
  
  if (rflags & REWRITE_FLAGS_VPATH_ONLY) {
    return cb(NULL, start_p, 0, cb_data);
  }
  else {
    return path_rewrite_1(start_p, cb, cb_data);
  }
}

void alloc_init(size_t max_balloc, size_t max_salloc)
{
  max_balloc = RECLEN(block_t) + ALIGN(max_balloc, _MAX_ALIGNMENT);
  max_salloc =
    RECLEN(block_t) + RECLEN(bstack_t) + ALIGN(max_salloc, _MAX_ALIGNMENT);
  block_size = MAX(max_balloc, max_salloc);
  mapping_size = RECLEN(mapping_t) + BLOCKS_PER_MAPPING * block_size;
  mapping_size = ALIGN(mapping_size, getpagesize());
}

void *block_alloc()
{
  mapping_t *mapping;
  block_t *block;
  
  // Search for a mapping with a free block or blocks
  mapping = first_mapping;
  while (mapping) {      
    if (mapping->in_use_bits != ALL_BLOCK_BITS) {
      int i;
      blocks_bits_t bits;
      
      // Search for a free block in mapping
      bits = ~mapping->in_use_bits;
      for (i = 0; i < BLOCKS_PER_MAPPING; i++) {
        if (bits & 1) {
          mapping->in_use_bits |= (blocks_bits_t)1 << i;
          block =
            (block_t *)((char *)HDR_TO(mapping_t, mapping) + i * block_size);
          block->block_idx = i;
          goto found_block;
        }
        
        bits >>= 1;
      }
    }
    
    mapping = mapping->next_mapping;
  }
  found_block:
  
  // Otherwise, allocate a new mapping
  if (!mapping) {
    mapping = (mapping_t *)mmap(
      NULL, mapping_size, PROT_READ | PROT_WRITE,
      MAP_PRIVATE | MAP_ANON, -1, 0);
    if (mapping == MAP_FAILED) {
      fprintf(stderr, "fsrewrite: mmap(2) failed.\n");
      abort();
    }
    
    // Add mapping to list
    if (first_mapping) {
      first_mapping->ptr_to_this = &mapping->next_mapping;
    }
    mapping->next_mapping = first_mapping;
    mapping->ptr_to_this = &first_mapping;
    first_mapping = mapping;
    
    // Initialize block
    mapping->in_use_bits = 1;
    block = (block_t *)HDR_TO(mapping_t, mapping);
    block->block_idx = 0;
  }
  
  return HDR_TO(block_t, block);
}

void block_free(void *ptr)
{
  mapping_t *mapping;
  block_t *block;
  
  block = TO_HDR(block_t, ptr);
  mapping = TO_HDR(mapping_t, (char *)block - (block->block_idx * block_size));
  mapping->in_use_bits &= ~((blocks_bits_t)1 << block->block_idx);
}

void *stack_alloc(void *ptr, size_t size)
{
  bstack_t *tail, *head;
  void *p;
  
  tail = head = NULL;
  if (ptr) {
    tail = TO_HDR(bstack_t, ptr);
    head = tail->head;
  }
  
  // Allocate a new block if necessary
  if (!tail ||
      (block_size - sizeof(block_t) - sizeof(bstack_t) - head->off < size)) {
    bstack_t *nhead;
    
    nhead = block_alloc();
    nhead->head = nhead;
    nhead->next = NULL;
    nhead->off = 0;
    
    if (tail) {
      head->next = nhead;
      tail->head = nhead;
    }
    
    head = nhead;
  }
  
  p = (char *)HDR_TO(bstack_t, head) + head->off;
  head->off += ALIGN(size, _MAX_ALIGNMENT);
  return p;
}

void stack_free(void *ptr)
{
  bstack_t *bstack, *next;
  
  if (!ptr) {
    return;
  }
  bstack = TO_HDR(bstack_t, ptr);
  while (bstack) {
    next = bstack->next;
    block_free(bstack);
    bstack = next;
  }
}

void *tree_op(
  tree_node_t **rootp, tree_cmp_cb_t cmp_cb, void *data, tree_action_t action)
{
  tree_node_t *root;
  
  root = *rootp;
  if (root) {
    int cmp = cmp_cb(data, HDR_TO(tree_node_t, root));
    if (cmp < 0) {
      return tree_op(&(root->left), cmp_cb, data, action);
    }
    else if (cmp > 0) {
      return tree_op(&(root->right), cmp_cb, data, action);
    }
    else {
      if (action == REMOVE) {
        if (root->left && !root->right) {
          *rootp = root->left;
        }
        else if (!root->left && root->right) {
          *rootp = root->right;
        }
        else if (root->left && root->right) {
          tree_node_t **temp;
          
          if (cmp & 1) {
            *rootp = root->left;
            temp = &root->left->right;
            while(*temp) {
              temp = &(*temp)->right;
            }
            *temp = root->right;
            
          }
          else {
            *rootp = root->right;
            temp = &root->right->left;
            while(*temp) {
              temp = &(*temp)->left;
            }
            *temp = root->left;
          }
        }
        else {
          *rootp = NULL;
        }
      }
      else if (action == TEST_INSERT) {
        return NULL;
      }
      return HDR_TO(tree_node_t, root);
    }
  }
  else {
    if (action == INSERT) {
      tree_node_t *node = TO_HDR(tree_node_t, data);
      
      node->left = NULL;
      node->right = NULL;
      node->next = NULL;
      *rootp = node;
    }
    else if (action == TEST_INSERT) {
      return rootp;
    }
    return NULL;
  }
}

static void tree_linearize_1(tree_node_t *node, tree_node_t ***prev_fref)
{
  if (node->left) {
    tree_linearize_1(node->left, prev_fref);
  }
  **prev_fref = node;
  *prev_fref = &(node->next);
  if (node->right) {
    tree_linearize_1(node->right, prev_fref);
  }
}

tree_node_t *tree_linearize(tree_node_t *node)
{
  tree_node_t *first, **prev_fref;
  
  first = NULL;
  prev_fref = &first;
  if (node) {
    tree_linearize_1(node, &prev_fref);
  }
  
  return first;
}
